/*
 Wireless Gameboy Controller - Arduino RX 
 Version: 1.0
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 25/05/2019
 Last Modified: 25/05/2019
 
 VCC - 3.3V, CE - Pin 2, CSN - Pin 10
 SCK - Pin 13, MI - Pin 12, MO - Pin 11
 
 */

#include <avr/io.h>
#include <util/delay.h>
#include <SPI.h>
#include "mirf.h"
#include "nRF24L01.h"

// Defines for setting the MiRF registers for transmitting or receiving mode
#define TX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )

uint8_t modeFound = 0;
uint8_t kpData = 0;
uint8_t kpDataHigh = 0;
uint8_t firstPacket = 1;
uint8_t volatile timedOut = 0;
uint8_t multiTxData[10];

// Write data using SPI
void spi_write_data(uint8_t * dataout, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    SPI.transfer(dataout[i]);
  }
}

// Read data using SPI
void spi_read_data(uint8_t * datain, uint8_t len) {
  uint8_t i;
  for (i = 0; i < len; i++) {
    datain[i] = SPI.transfer(0x00);
  }
}

// Initializes pins ans interrupt to communicate with the MiRF module
// Should be called in the early initializing phase at startup.
void mirf_init(void) {
  // Define CSN and CE as Output and set them to default
  pinMode (CSN, OUTPUT);
  pinMode (CE, OUTPUT);

  mirf_CE_lo;
  mirf_CSN_hi;
}

// Sets the important registers in the MiRF module and powers the module
// in receiving mode
void mirf_config(void) {
  // Set RF channel  
  mirf_config_register(RF_CH, nrfChannel);
  
  // Set length of incoming payload 
  mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
  
  // Disable auto acknowledgement
  mirf_config_register(EN_AA, 0<<ENAA_P0);
  
  // Disable auto retransmit delay
  mirf_config_register(SETUP_RETR, 0);
  
  // Set RADDR and TADDR
  mirf_write_register(RX_ADDR_P0, nrfRx, 5);
  mirf_write_register(TX_ADDR, TADDR, 5);
  
  // Enable RX_ADDR_P0 address matching
  mirf_config_register(EN_RXADDR, (1<<ERX_P0));
}

// Set registers in the nRF module for updating the RX settings
void mirf_config_setting_update(void) {
  // Set RF channel 
  mirf_config_register(RF_CH, mirf_CH_SETTING_UPDATE);
  
  // Set length of incoming payload 
  mirf_config_register(RX_PW_P0, mirf_PAYLOAD_SETTING_UPDATE);
  
  // Set RADDR
  mirf_write_register(RX_ADDR_P0, RADDR_SETTING_UPDATE, 5);
}

// Flush RX and TX FIFO
void mirf_flush_rx_tx(void) {
  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(FLUSH_RX); // Flush RX
  mirf_CSN_hi; // Pull up chip select

  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(FLUSH_TX);  // Write cmd to flush tx fifo
  mirf_CSN_hi; // Pull up chip select
}

// Read the status register
uint8_t mirf_status(void) {
  mirf_CSN_lo; // Pull down chip select
  SPI.transfer(R_REGISTER | (REGISTER_MASK & STATUS));
  uint8_t status = SPI.transfer(NOP); // Read status register
  mirf_CSN_hi; // Pull up chip select
  return status;
}

// Write one byte into the given MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  SPI.transfer(value);
  mirf_CSN_hi;
}

// Reads an array of bytes from the given start position in the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
  spi_read_data(value, len);
  mirf_CSN_hi;
}

// Writes an array of bytes into inte the MiRF registers
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
  mirf_CSN_lo;
  SPI.transfer(W_REGISTER | (REGISTER_MASK & reg));
  spi_write_data(value, len);
  mirf_CSN_hi;
}

uint8_t mirf_read_reg(uint8_t reg) {
  mirf_CSN_lo;
  SPI.transfer(R_REGISTER | (REGISTER_MASK & reg));
  uint8_t value = SPI.transfer(0x00);
  mirf_CSN_hi;
  return value;
}

// Check to see if the TX is sending an updated RX channel/address to use
void check_for_new_settings(void) {
  mirf_config_setting_update();
  
  RX_POWERUP;
  _delay_ms(5);
  mirf_flush_rx_tx(); // Flush TX/RX
  mirf_CE_hi; // Start listening
  
  // Check for changing of our RX channel/address for 0.5 seconds
  for (uint8_t timer = 0; timer < 10; timer++) {
    if (mirf_status() & (1<<RX_DR)) { 
      mirf_CE_lo; // Stop listening
      
      mirf_CSN_lo; // Pull down chip select
      SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
      nrfChannel = SPI.transfer(0); // Read data
      nrfRx[4] = SPI.transfer(0);
      mirf_CSN_hi; // Pull up chip select
      mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
      
      mirf_flush_rx_tx(); // Flush TX/RX
      
      // Update our EEPROM
      eeprom_write_byte((uint8_t*) nrfChannelLocation, nrfChannel);
      eeprom_write_byte((uint8_t*) nrfRxLocation, nrfRx[4]);
      
      _delay_ms(250);
      //digitalWrite(LED_BUILTIN, HIGH);
      _delay_ms(250);
      //digitalWrite(LED_BUILTIN, LOW);
      
      break;
    }
    
    _delay_ms(50);
  }
  mirf_CE_lo;
  
  mirf_config(); // Reconfigure again
}

// Wait for a transmitter to connect
void check_for_tx_mode(void) {
  modeFound = 0;
  
  RX_POWERUP;
  _delay_ms(5);
  mirf_flush_rx_tx(); // Flush TX/RX
  mirf_CE_hi; // Start listening

  Serial.print("Waiting for signal");
  
  // Set payload size to 9 bytes (will accept any packet length from 1 to 9 bytes). 
  // Once a packet is received, read it and check for the data it contains
  uint8_t payloadSize = 9;
  while (modeFound == 0) {
    mirf_config_register(RX_PW_P0, payloadSize);
    mirf_flush_rx_tx(); // Flush TX/RX
    mirf_CE_hi; // Start listening
    
    for (uint8_t timer = 0; timer < 50; timer++) {
      if (mirf_status() & (1<<RX_DR)) { 
        mirf_CE_lo; // Stop listening
        mirf_CSN_lo; // Pull down chip select
        SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
        
        // Read out
        for (uint8_t x = 0; x < payloadSize; x++) {
          multiTxData[x] = SPI.transfer(0);
        }
        mirf_CSN_hi; // Pull up chip select
        mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
        mirf_flush_rx_tx();
        
        // Check mode
        if (multiTxData[0] == SNES_TX) {
          modeFound = SNES_TX;
          payloadSize = 3;
          mirf_config_register(RX_PW_P0, payloadSize);
        }
        else if (multiTxData[1] != 0xBF) { // GBA mode
          modeFound = GBA_MODE;
          payloadSize = 2;
          mirf_config_register(RX_PW_P0, payloadSize);
        }
        else if (multiTxData[1] == 0xBF) { // GB mode (0xBF means no data in that byte)
          modeFound = GB_MODE;
          payloadSize = 1;
          mirf_config_register(RX_PW_P0, payloadSize);
        }
        
        break;
      }
      
      _delay_ms(10);
    }
    mirf_CE_lo;
    
    //digitalWrite(LED_BUILTIN, HIGH);
    _delay_ms(5);
    //digitalWrite(LED_BUILTIN, LOW);
    
    if (modeFound >= 1) {
      break;
    }

    Serial.print(".");
  }
  
  RX_POWERUP;
  _delay_ms(5);
  mirf_flush_rx_tx(); // Flush TX/RX
  mirf_CE_hi; // Start listening
}

void setup() {
  Serial.begin(9600);
  SPI.begin();
  mirf_init();
  _delay_ms(50);
  mirf_config();
  Serial.println("nRF24 Configured");
  
  // Load nRF channel and address from EEPROM
  /*nrfChannel = eeprom_read_byte((uint8_t*) nrfChannelLocation);
  nrfRx[4] = eeprom_read_byte((uint8_t*) nrfRxLocation);
  if (nrfChannel == blankEEPROMByte) {
    nrfChannel = 1;
    nrfRx[4] = 1;
  }*/

  // Set to channel 1 and address 1
  nrfChannel = 1;
  nrfRx[4] = 1;
}

void loop() {
  check_for_new_settings(); // Check for new nRF settings
  check_for_tx_mode(); // Check for TX mode
  
  while(1) {
    // Packet received
    if (mirf_status() & (1<<RX_DR)) { // Packet received
      mirf_CE_lo; // Stop listening
      
      mirf_CSN_lo; // Pull down chip select
      SPI.transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
      
      if (modeFound == GBA_MODE) { // Load 2 bytes if in GBA mode
        kpDataHigh = SPI.transfer(0);
        kpData = SPI.transfer(0);
      }
      else {
        kpData = SPI.transfer(0); // Read data
      }
      
      mirf_CSN_hi; // Pull up chip select
      mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
      mirf_flush_rx_tx(); // Flush TX/RX
      mirf_CE_hi; // Start listening
      //digitalWrite(LED_BUILTIN, LOW);
      
      // Process keys pressed
      if (kpData >= 1 || kpDataHigh >= 1) { 
        Serial.print("Keys pressed = ");
        if (kpData & 0x01) { // A
          Serial.print("A ");
        }
        if (kpData & 0x02) { // B
          Serial.print("B ");
        }
        if (kpData & 0x04) { // Right
          Serial.print("Right ");
        }
        if (kpData & 0x08) { // Left
          Serial.print("Left ");
        }
        if (kpData & 0x10) { // Up
          Serial.print("Up ");
        }
        if (kpData & 0x20) { // Down
          Serial.print("Down ");
        }
        if (kpData & 0x40) { // Select
          Serial.print("Select ");
        }
        if (kpData & 0x80) { // Start
          Serial.print("Start ");
        }
        if (kpDataHigh & 0x01) { // L (for GBA)
          Serial.print("L ");
        }
        if (kpDataHigh & 0x02) { // R (for GBA)
          Serial.print("R ");
        }
        Serial.println("");
      }
    }
    
    _delay_us(250);
  }
}
